@@ -18,6 +18,7 @@
 #include <cstdio>  // std::fgets(), std::FILE
 #include <cerrno>  // ENXIO
 #include <set>
+#include <map>
 #include <vector>
 #include <utility>  // std::pair
 
@@ -123,10 +124,28 @@ inline std::string detect_drives_linux_udev_byid(std::vector<std::string>& devic
 
 
 
+/// Cache of file->contents, caches read files.
+std::map<std::string, std::string> s_read_file_cache;
+
+
+
+/// Clear the read file cache.
+inline void clear_read_file_cache()
+{
+	s_read_file_cache.clear();
+}
+
+
+
 /// Procfs files don't support SEEK_END or ftello() (I think). They can't
 /// be read through hz::File::get_contents, so use this function instead.
-inline bool read_proc_file(hz::File& file, std::vector<std::string>& lines)
+inline bool read_proc_file(hz::File& file, std::string& contents)
 {
+	if (s_read_file_cache.find(file.get_path()) != s_read_file_cache.end()) {
+		contents = s_read_file_cache[file.get_path()];
+		return true;
+	}
+
 	if (!file.open("rb"))  // closed automatically
 		return false;  // the error message is in File itself.
 
@@ -137,26 +156,34 @@ inline bool read_proc_file(hz::File& file, std::vector<std::string>& lines)
 	char line[256];
 	while (std::fgets(line, static_cast<int>(sizeof(line)), fp) != NULL) {
 		if (*line != '\0')
-			lines.push_back(line);
+			contents += line;  // line contains the terminating newline as well
 	}
 
+	s_read_file_cache[file.get_path()] = contents;
+
+	debug_begin();  // avoiding printing prefix on every line
+	debug_out_dump("app", DBG_FUNC_MSG << "File contents (\"" << file.get_path() << "\"):\n" << contents << "\n");
+	debug_end();
+
 	return true;
 }
 
 
 
-/// Read a file completely from /proc
-inline bool read_proc_complete_file(hz::File& file, std::string& contents)
+/// Same as read_proc_file(), but returns a vector of lines
+inline bool read_proc_file_lines(hz::File& file, std::vector<std::string>& lines)
 {
-	// The crude way
-	std::vector<std::string> lines;
-	bool status = read_proc_file(file, lines);
-	contents = hz::string_join(lines, '\n');
+	std::string contents;
+	bool status = read_proc_file(file, contents);
+	if (status) {
+		hz::string_split(contents, '\n', lines, true);
+	}
 	return status;
 }
 
 
 
+
 /// Read /proc/partitions file. Return error message on error.
 inline std::string read_proc_partitions_file(std::vector<std::string>& lines)
 {
@@ -167,7 +194,7 @@ inline std::string read_proc_partitions_file(std::vector<std::string>& lines)
 	}
 
 	hz::File file(path);
-	if (!read_proc_file(file, lines)) {  // this outputs to debug too
+	if (!read_proc_file_lines(file, lines)) {  // this outputs to debug too
 		std::string error_msg = file.get_error_utf8();  // save before calling other file functions
 		if (!file.exists()) {
 			debug_out_warn("app", DBG_FUNC_MSG << "Partitions file doesn't exist.\n");
@@ -192,7 +219,7 @@ inline std::string read_proc_devices_file(std::vector<std::string>& lines)
 	}
 
 	hz::File file(path);
-	if (!read_proc_file(file, lines)) {  // this outputs to debug too
+	if (!read_proc_file_lines(file, lines)) {  // this outputs to debug too
 		std::string error_msg = file.get_error_utf8();  // save before calling other file functions
 		if (!file.exists()) {
 			debug_out_warn("app", DBG_FUNC_MSG << "Devices file doesn't exist.\n");
@@ -221,7 +248,7 @@ inline std::string read_proc_scsi_scsi_file(std::vector< std::pair<int, std::str
 	hz::File file(path);
 	std::vector<std::string> lines;
 
-	if (!read_proc_file(file, lines)) {  // this outputs to debug too
+	if (!read_proc_file_lines(file, lines)) {  // this outputs to debug too
 		std::string error_msg = file.get_error_utf8();  // save before calling other file functions
 		if (!file.exists()) {
 			debug_out_warn("app", DBG_FUNC_MSG << "SCSI file doesn't exist.\n");
@@ -269,7 +296,7 @@ inline std::string read_proc_scsi_sg_devices_file(std::vector< std::vector<int>
 	hz::File file(path);
 	std::vector<std::string> lines;
 
-	if (!read_proc_file(file, lines)) {  // this outputs to debug too
+	if (!read_proc_file_lines(file, lines)) {  // this outputs to debug too
 		std::string error_msg = file.get_error_utf8();  // save before calling other file functions
 		if (!file.exists()) {
 			debug_out_warn("app", DBG_FUNC_MSG << "Sg devices file doesn't exist.\n");
@@ -302,38 +329,6 @@ inline std::string read_proc_scsi_sg_devices_file(std::vector< std::vector<int>
 
 
 
-/// Get number of ports by sequentially running smartctl on each port, until
-/// one of the gives an error. Return -1 on error.
-inline std::string smartctl_get_drives(const std::string& dev, const std::string& type,
-	  int from, int to, std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
-{
-	hz::intrusive_ptr<CmdexSync> smartctl_ex = ex_factory->create_executor(ExecutorFactory::ExecutorSmartctl);
-
-	for (int i = from; i <= to; ++i) {
-		std::string type_arg = hz::string_sprintf(type.c_str(), i);
-		StorageDeviceRefPtr drive(new StorageDevice(dev, type_arg));
-
-// 		std::string error_msg = execute_smartctl(dev, "-d " + type_arg, "-i", smartctl_ex, output);
-		std::string error_msg = drive->fetch_basic_data_and_parse(smartctl_ex);
-		std::string output = drive->get_info_output();
-
-		// if we've reached smartctl port limit (older versions may have smaller limits), abort.
-		if (app_pcre_match("/VALID ARGUMENTS ARE/mi", output)) {
-			break;
-		}
-
-		if (!error_msg.empty()) {
-			debug_out_info("app", "Smartctl returned with an error: " << error_msg << "\n");
-		} else {
-			drives.push_back(drive);
-		}
-	}
-
-	return std::string();
-}
-
-
-
 
 /**
 <pre>
@@ -383,7 +378,7 @@ major minor  #blocks  name
 </pre> */
 inline std::string detect_drives_linux_proc_partitions(std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
 {
-	debug_out_info("app", DBG_FUNC_MSG << "Detecting through /proc/partitions...\n");
+	debug_out_info("app", DBG_FUNC_MSG << "Detecting drives through partitions file (/proc/partitions by default; set \"system/linux_proc_partitions_path\" config key to override).\n");
 
 	std::vector<std::string> lines;
 	std::string error_msg = read_proc_partitions_file(lines);
@@ -392,7 +387,6 @@ inline std::string detect_drives_linux_proc_partitions(std::vector<StorageDevice
 	}
 
 	std::vector<std::string> blacklist;
-	// fixme: not sure about how partitions are visible with twa0.
 	blacklist.push_back("/d[a-z][0-9]+$/");  // sda1, hdb2 - partitions. twa0 and twe1 are drives, not partitions.
 	blacklist.push_back("/ram[0-9]+$/");  // ramdisks?
 	blacklist.push_back("/loop[0-9]*$/");  // not sure if loop devices go there, but anyway...
@@ -426,8 +420,9 @@ inline std::string detect_drives_linux_proc_partitions(std::vector<StorageDevice
 			continue;
 
 		std::string path = "/dev/" + dev;  // let's just hope the it's /dev.
-		if (std::find(devices.begin(), devices.end(), path) == devices.end())  // there may be duplicates
+		if (std::find(devices.begin(), devices.end(), path) == devices.end()) {  // there may be duplicates
 			devices.push_back(path);
+		}
 	}
 
 	hz::intrusive_ptr<CmdexSync> smartctl_ex = ex_factory->create_executor(ExecutorFactory::ExecutorSmartctl);
@@ -439,8 +434,11 @@ inline std::string detect_drives_linux_proc_partitions(std::vector<StorageDevice
 		// 3ware controllers also export themselves as sd*. Smartctl detects that,
 		// so we can avoid adding them. Older smartctl (5.38) prints "AMCC", newer one
 		// prints "AMCC/3ware controller". It's better to search it this way.
-		if (!app_pcre_match("/try adding '-d 3ware,N'/im", drive->get_info_output())) {
+		if (app_pcre_match("/try adding '-d 3ware,N'/im", drive->get_info_output())) {
+			debug_out_dump("app", "Drive " << drive->get_device_with_type() << " seems to be a 3ware controller, ignoring.\n");
+		} else {
 			drives.push_back(drive);
+			debug_out_info("app", "Added drive " << drive->get_device_with_type() << ".\n");
 		}
 	}
 
@@ -478,6 +476,8 @@ how they will be ordered for tw_cli.
 </pre> */
 inline std::string detect_drives_linux_3ware(std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
 {
+	debug_out_info("app", DBG_FUNC_MSG << "Detecting drives behind 3ware controller(s)...\n");
+
 	std::vector<std::string> lines;
 	std::string error_msg = read_proc_devices_file(lines);
 	if (!error_msg.empty()) {
@@ -492,7 +492,7 @@ inline std::string detect_drives_linux_3ware(std::vector<StorageDeviceRefPtr>& d
 	for (std::size_t i = 0; i < lines.size(); ++i) {
 		std::string dev;
 		if (app_pcre_match("/^[ \\t]*[0-9]+[ \\t]+(tw[ael])(?:[ \\t]*|$)/", hz::string_trim_copy(lines[i]), &dev)) {
-			debug_out_dump("app", DBG_FUNC_MSG << "Found " << dev << " entry in devices file.\n");
+			debug_out_dump("app", DBG_FUNC_MSG << "Found 3ware " << dev << " entry in devices file.\n");
 			if (dev == "twa") {
 				twa_found = true;
 			} else if (dev == "twe") {
@@ -506,11 +506,13 @@ inline std::string detect_drives_linux_3ware(std::vector<StorageDeviceRefPtr>& d
 	}
 
 	if (!twa_found && !twe_found && !twl_found) {
+		debug_out_info("app", DBG_FUNC_MSG << "No 3ware-specific entries found in devices file.\n");
 		return std::string();  // no controllers
 	}
 
 	lines.clear();
 
+	debug_out_dump("app", DBG_FUNC_MSG << "Checking scsi file for 3ware controllers.\n");
 	std::vector< std::pair<int, std::string> > vendors_models;
 	error_msg = read_proc_scsi_scsi_file(vendors_models);
 	if (!error_msg.empty()) {
@@ -518,7 +520,7 @@ inline std::string detect_drives_linux_3ware(std::vector<StorageDeviceRefPtr>& d
 	}
 
 
-	std::set<int> controller_hosts;  // scsi hosts found for tw*
+	std::set<int> controller_hosts;  // scsi hosts (controllers) found for tw*
 	std::map<std::string, int> device_numbers;  // device base (e.g. twa) -> number of times found
 
 	for (std::size_t i = 0; i < vendors_models.size(); ++i) {
@@ -566,9 +568,13 @@ inline std::string detect_drives_linux_3ware(std::vector<StorageDeviceRefPtr>& d
 
 		error_msg = tw_cli_get_drives(dev, vendors_models[i].first, drives, ex_factory, false);
 		if (!error_msg.empty()) {  // no tw_cli
-			int max_ports = rconfig::get_data<int32_t>("system/linux_max_scan_ports");
-			max_ports = std::max(max_ports, 23);  // 128 smartctl calls are too much (it's too slow). Settle for 24.
-			error_msg = smartctl_get_drives(dev, "3ware,%d", 0, max_ports, drives, ex_factory);
+			int max_ports = rconfig::get_data<int32_t>("system/linux_3ware_max_scan_port");
+			max_ports = std::max(0, std::min(max_ports, 127));  // Sanity check
+			debug_out_dump("app", "Starting brute-force port scan on 0-" << max_ports << " ports, device \"" << dev
+					<< "\". Change the maximum by setting \"system/linux_3ware_max_scan_port\" config key.\n");
+			std::string last_output;
+			error_msg = smartctl_scan_drives_sequentially(dev, "3ware,%d", 0, max_ports, drives, ex_factory, last_output);
+			debug_out_dump("app", "Brute-force port scan finished.\n");
 		}
 
 		if (!error_msg.empty()) {
@@ -604,6 +610,8 @@ sure how to detect the failure), fall back to "-d scsi".
 </pre> */
 inline std::string detect_drives_linux_adaptec(std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
 {
+	debug_out_info("app", DBG_FUNC_MSG << "Detecting drives behind Adaptec controller(s)...\n");
+
 	std::vector<std::string> lines;
 	std::string error_msg = read_proc_devices_file(lines);
 	if (!error_msg.empty()) {
@@ -612,7 +620,7 @@ inline std::string detect_drives_linux_adaptec(std::vector<StorageDeviceRefPtr>&
 
 	bool aac_found = false;
 
-	// Check /proc/devices for twa or twe
+	// Check /proc/devices for "aac"
 	for (std::size_t i = 0; i < lines.size(); ++i) {
 		if (app_pcre_match("/^[ \\t]*[0-9]+[ \\t]+aac(?:[ \\t]*|$)/", hz::string_trim_copy(lines[i]))) {
 			debug_out_dump("app", DBG_FUNC_MSG << "Found aac entry in devices file.\n");
@@ -620,11 +628,13 @@ inline std::string detect_drives_linux_adaptec(std::vector<StorageDeviceRefPtr>&
 		}
 	}
 	if (!aac_found) {
+		debug_out_info("app", DBG_FUNC_MSG << "No Adaptec-specific entries found in devices file.\n");
 		return std::string();  // no controllers
 	}
 
 	lines.clear();
 
+	debug_out_dump("app", DBG_FUNC_MSG << "Checking scsi file for Adaptec controllers.\n");
 	std::vector< std::pair<int, std::string> > vendors_models;
 	error_msg = read_proc_scsi_scsi_file(vendors_models);
 	if (!error_msg.empty()) {
@@ -683,8 +693,10 @@ inline std::string detect_drives_linux_adaptec(std::vector<StorageDeviceRefPtr>&
 
 			if (!local_error_msg.empty()) {
 				debug_out_info("app", "Smartctl returned with an error: " << local_error_msg << "\n");
+				debug_out_dump("app", "Skipping drive " << drive->get_device_with_type() << ".\n");
 			} else {
 				drives.push_back(drive);
+				debug_out_info("app", "Added drive " << drive->get_device_with_type() << ".\n");
 			}
 		}
 	}
@@ -698,9 +710,15 @@ inline std::string detect_drives_linux_adaptec(std::vector<StorageDeviceRefPtr>&
 
 
 
-/** </tt>
+/** <pre>
 Areca Linux (arcmsr driver):
-Call as: smartctl -i -d areca,[1-24] /dev/sg0
+Call as:
+	- smartctl -i -d areca,N /dev/sg0  (N is [1,24]).
+	- smartctl -i -d areca,N/E /dev/sg0  (N is [1,128], E is [1,8]).
+The models that have "ix" (case-insensitive) in their model names
+have the enclosures and require the N,E syntax.
+Note: Using N/E syntax with non-enclosure cards seems to work
+regardless of the value of E.
 
 Detection:
 	There don't seem to be any entries in /proc/devices.
@@ -712,25 +730,29 @@ Detect SCSI host (N in hostN below):
 	The line number in /proc/scsi/sg/devices is X in /dev/sgX.
 Check /sys/bus/scsi/devices/hostN/scsi_host/hostN/host_fw_hd_channels, set
 	its contents as the number of ports. If not present, use 24 (maximum).
-Probe each port for a valid drive: If the output contains "Device Read Identity Failed",
-	then there is no drive on that port (or Areca has an old firmware, nothing we can
-	do there).
+	No-enclosure models only, since we have no info yet for the enclosure ones.
+Probe each port for a valid drive: If the output contains "Device Read Identity Failed"
+	or "empty IDENTIFY data", then there is no drive on that port (or Areca has an
+	old firmware, nothing we can do there).
 Notification: If /sys/bus/scsi/devices/hostN/scsi_host/hostN/host_fw_version
-	is older than "V1.46 2009-01-06", notify the user (maybe its better to grep
-	the smartctl output for that on port 0?). NOT IMPLEMENTED YET.
-</tt> */
+	is older than "V1.46 2009-01-06", (1.51 for enclosure-having cards) notify the user
+	(maybe its better to grep the smartctl output for that on port 0?). NOT IMPLEMENTED YET.
+</pre> */
 inline std::string detect_drives_linux_areca(std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
 {
+	debug_out_info("app", DBG_FUNC_MSG << "Detecting drives behind Areca controller(s)...\n");
+
 	std::vector< std::pair<int, std::string> > vendors_models;
 	std::string error_msg = read_proc_scsi_scsi_file(vendors_models);
 	if (!error_msg.empty()) {
 		return error_msg;
 	}
 
-	std::set<int> controller_hosts;
+	std::map<int, bool> controller_hosts;  // controller # -> has_enclosure
 
 	for (std::size_t i = 0; i < vendors_models.size(); ++i) {
-		if (!app_pcre_match("/Vendor: Areca /i", vendors_models[i].second)) {
+		std::string vendor_line = vendors_models[i].second;
+		if (!app_pcre_match("/Vendor: Areca /i", vendor_line)) {
 			continue;  // not a supported controller
 		}
 		int host_num = vendors_models[i].first;
@@ -741,10 +763,18 @@ inline std::string detect_drives_linux_areca(std::vector<StorageDeviceRefPtr>& d
 			debug_out_dump("app", "Skipping adapter with SCSI host " << host_num << ", host already found.\n");
 			continue;
 		}
-		controller_hosts.insert(host_num);
+
+		// Check if it contains the "ix" string in its model name. It may not be exactly
+		// a suffix (there may be "-VOL#00" or something like that appended as well).
+		bool has_enclosure = app_pcre_match("/Model:.+ix.+Rev:/i", vendor_line);
+		debug_out_dump("app", DBG_FUNC_MSG << "Areca controller (SCSI host " << host_num
+				<< ") seems to " << (has_enclosure ? "have enclosure(s)" : "have no enclosures") << ".\n");
+
+		controller_hosts[host_num] = has_enclosure;
 	}
 
 	if (controller_hosts.empty()) {
+		debug_out_info("app", DBG_FUNC_MSG << "No Areca-specific entries found in SCSI file.\n");
 		return error_msg;
 	}
 
@@ -756,8 +786,9 @@ inline std::string detect_drives_linux_areca(std::vector<StorageDeviceRefPtr>& d
 
 	hz::intrusive_ptr<CmdexSync> smartctl_ex = ex_factory->create_executor(ExecutorFactory::ExecutorSmartctl);
 
-	for (std::set<int>::iterator iter = controller_hosts.begin(); iter != controller_hosts.end(); ++iter) {
-		const int host_num = *iter;
+	for (std::map<int, bool>::iterator iter = controller_hosts.begin(); iter != controller_hosts.end(); ++iter) {
+		const int host_num = iter->first;
+		const bool has_enclosure = iter->second;
 
 		for (std::size_t sg_num = 0; sg_num < sg_entries.size(); ++sg_num) {
 			if (sg_entries[sg_num].size() < 5) {
@@ -770,26 +801,53 @@ inline std::string detect_drives_linux_areca(std::vector<StorageDeviceRefPtr>& d
 				continue;
 			}
 
-			int max_ports = 0;
+			if (has_enclosure) {
+				// TODO We have no information on what "/sys/bus/scsi/devices/host%d/scsi_host/host%d/host_fw_hd_channels"
+				// contains in case of enclosure-having cards.
+
+				int max_enclosures = rconfig::get_data<int32_t>("system/linux_areca_enc_max_enclosure");
+				max_enclosures = std::max(1, std::min(8, max_enclosures));  // 1-8
+
+				int max_ports = rconfig::get_data<int32_t>("system/linux_areca_enc_max_scan_port");
+				max_ports = std::max(1, std::min(128, max_ports));  // 1-128 sanity check
+
+				std::string dev = std::string("/dev/sg") + hz::number_to_string(sg_num);
+
+				debug_out_dump("app", "Starting brute-force port/enclosure scan on 1-" << max_ports << " ports, 1-" << max_enclosures << " enclosures, device \"" << dev
+						<< "\". Change the maximums by setting \"system/linux_areca_enc_max_scan_port\" and \"system/linux_areca_enc_max_enclosure\" config keys.\n");
+				std::string last_output;
+				for (int enclosure_no = 1; enclosure_no < max_enclosures; ++enclosure_no) {
+					error_msg = smartctl_scan_drives_sequentially(dev, "areca,%d/" + hz::number_to_string(enclosure_no), 1, max_ports, drives, ex_factory, last_output);
+				}
+				debug_out_dump("app", "Brute-force port/enclosure scan finished.\n");
 
-			std::string ports_path = hz::string_sprintf("/sys/bus/scsi/devices/host%d/scsi_host/host%d/host_fw_hd_channels", host_num, host_num);
-			hz::File ports_file(ports_path);
-			std::string ports_file_contents;
-			if (!read_proc_complete_file(ports_file, ports_file_contents)) {
-				debug_out_warn("app", DBG_FUNC_MSG << "Couldn't read number of ports on Areca controller: "
-						<< ports_file.get_error_utf8() << ", trying manually.\n");
 			} else {
-				hz::string_is_numeric(hz::string_trim_copy(ports_file_contents), max_ports);
+				int max_ports = 0;
+
+				// Read the number of ports.
+				std::string ports_path = hz::string_sprintf("/sys/bus/scsi/devices/host%d/scsi_host/host%d/host_fw_hd_channels", host_num, host_num);
+				hz::File ports_file(ports_path);
+				std::string ports_file_contents;
+				if (!read_proc_file(ports_file, ports_file_contents)) {
+					debug_out_warn("app", DBG_FUNC_MSG << "Couldn't read the number of ports on Areca controller (\"" << ports_path << "\"): "
+							<< ports_file.get_error_utf8() << ", trying manually.\n");
+				} else {
+					hz::string_is_numeric(hz::string_trim_copy(ports_file_contents), max_ports);
+					debug_out_dump("app", DBG_FUNC_MSG << "Detected " << max_ports << " ports, through \"" << ports_path << "\".\n");
+				}
+				if (max_ports == 0) {
+					max_ports = rconfig::get_data<int32_t>("system/linux_areca_noenc_max_scan_port");
+				}
+				max_ports = std::max(1, std::min(24, max_ports));  // 1-24 sanity check
+
+				std::string dev = std::string("/dev/sg") + hz::number_to_string(sg_num);
+				debug_out_dump("app", "Starting brute-force port scan on 1-" << max_ports << " ports, device \"" << dev
+						<< "\". Change the maximum by setting \"system/linux_areca_noenc_max_scan_port\" config key.\n");
+				std::string last_output;
+				error_msg = smartctl_scan_drives_sequentially(dev, "areca,%d", 1, max_ports, drives, ex_factory, last_output);
+				debug_out_dump("app", "Brute-force port scan finished.\n");
 			}
 
-			if (max_ports < 1 || max_ports > 24) {
-				max_ports = 24;  // default to 24 ports (smartctl maximum for Areca)
-			}
-
-			std::string dev = std::string("/dev/sg") + hz::number_to_string(sg_num);
-
-			error_msg = smartctl_get_drives(dev, "areca,%d", 1, max_ports, drives, ex_factory);
-
 			if (!error_msg.empty()) {
 				debug_out_warn("app", DBG_FUNC_MSG << "Couldn't get the drives on ports of Areca controller: " << error_msg << "\n");
 			}
@@ -801,6 +859,211 @@ inline std::string detect_drives_linux_areca(std::vector<StorageDeviceRefPtr>& d
 
 
 
+/** <pre>
+Detect drives behind HP RAID controller (cciss driver).
+(aka CCISS (HP (Compaq) Smart Array Controller).
+
+Note: hpsa/hpahcisr have a different method of detection.
+
+Call as: smartctl -a -d cciss,[0-127] /dev/cciss/c0d0
+	With the cciss drivers , for the path /dev/cciss/c0d0, c0 means controler 0 and d0 means LUN 1.
+	So smarctl provides the same values for c0d0 and c0d1. There are just two logical drives
+	on the same Smart Array controler.
+To detect controller presence: cat /proc/devices, contains "cciss0"
+	for controller 0.
+	/proc/scsi/scsi contains no entries for it.
+/dev/cciss/c0d0p1 is the first ordinary partition (used in mount, etc...).
+
+The port limit had a brief regression in 5.39.x (limited to 15).
+5.39 doesn't seem to work at all (prereleases work with P400 controller)?
+	Possibly a 64-bit-only issue?
+There is a cciss_vol_status utility but it doesn't report anything except "OK" status.
+It may not return the "SMART Health Status: OK" line in smartctl -i for some reason?
+	(It returns it in -a though, and in some freebsd output I found).
+If no /dev/cciss exists, you may need to run "cd /dev; ./MAKEDEV cciss".
+
+Detection:
+	Grep /proc/devices for "cciss([0-9]+)" where the number is the controller #.
+	Run "smartctl -i -d cciss,[0-127] /dev/cciss/cNd0" (where N is the controller #),
+		until "No such device or address" or "VALID ARGUMENTS ARE" is encountered in the output.
+	Note: We're not sure how to differentiate the outputs of free / non-existent ports,
+		so scan them until 15, just in case.
+</pre> */
+inline std::string detect_drives_linux_cciss(std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
+{
+	debug_out_info("app", DBG_FUNC_MSG << "Detecting drives behind HP RAID (CCISS) controller(s)...\n");
+
+	std::vector<std::string> lines;
+	std::string error_msg = read_proc_devices_file(lines);
+	if (!error_msg.empty()) {
+		return error_msg;
+	}
+
+	std::vector<int> controllers;
+
+	// Check /proc/devices for ccissX (where X is the controller #).
+	for (std::size_t i = 0; i < lines.size(); ++i) {
+		std::string controller_no_str;
+		if (app_pcre_match("/^[ \\t]*[0-9]+[ \\t]+cciss([0-9]+)(?:[ \\t]*|$)/", hz::string_trim_copy(lines[i]), &controller_no_str)) {
+			debug_out_dump("app", DBG_FUNC_MSG << "Found cciss" << controller_no_str << " entry in devices file.\n");
+			int controller_no = 0;
+			hz::string_is_numeric(controller_no_str, controller_no);
+			controllers.push_back(controller_no);
+		}
+	}
+	if (controllers.empty()) {
+		debug_out_info("app", DBG_FUNC_MSG << "No cciss-specific entries found in devices file.\n");
+		return std::string();  // no controllers
+	}
+
+	hz::intrusive_ptr<CmdexSync> smartctl_ex = ex_factory->create_executor(ExecutorFactory::ExecutorSmartctl);
+
+	for (std::size_t controller_index = 0; controller_index < controllers.size(); ++controller_index) {
+		int controller_no = controllers.at(controller_index);
+
+		std::string dev = std::string("/dev/cciss/c") + hz::number_to_string(controller_no) + "d0";
+
+		const int max_port = 127;
+		debug_out_dump("app", "Starting brute-force port scan on 1-" << max_port << " ports, device \"" << dev << "\".\n");
+
+		for (int port = 0; port <= max_port; ++port) {
+			StorageDeviceRefPtr drive(new StorageDevice(dev, std::string("cciss,") + hz::number_to_string(port)));
+
+			std::string local_error_msg = drive->fetch_basic_data_and_parse(smartctl_ex);
+			std::string output = drive->get_info_output();
+
+			if (!local_error_msg.empty()) {
+				debug_out_info("app", "Smartctl returned with an error: " << local_error_msg << "\n");
+			}
+
+			if (app_pcre_match("/VALID ARGUMENTS ARE/mi", output)) {
+				// smartctl doesn't support this many ports, return.
+				debug_out_dump("app", "Reached smartctl port limit with port " << port << ", stopping port scan.\n");
+				break;
+			}
+			if (app_pcre_match("/No such device or address/mi", output) && port > 15) {
+				// we've reached the controller port limit
+				debug_out_dump("app", "Reached controller port limit with port " << port << ", stopping port scan.\n");
+				break;
+			}
+
+			if (local_error_msg.empty()) {
+				drives.push_back(drive);
+				debug_out_info("app", "Added drive " << drive->get_device_with_type() << ".\n");
+			} else {
+				debug_out_dump("app", "Skipping drive " << drive->get_device_with_type() << " due to smartctl error.\n");
+			}
+		}
+
+		debug_out_dump("app", "Brute-force port scan finished.\n");
+	}
+
+	return error_msg;
+}
+
+
+
+/** <pre>
+Detect drives behind HP RAID controller (hpsa / hpahcisr drivers).
+
+May need "-d sat+cciss,N" for SATA, but is usually auto-detected.
+
+smartctl -a -d cciss,0 /dev/sg2    (hpsa or hpahcisr drivers under Linux)
+("lsscsi -g" is helpful in determining which scsi generic device node corresponds to which device.)
+Use the nodes corresponding to the RAID controllers, not the nodes corresponding to logical drives.
+No entry in /proc/devices.
+/dev/sda also seems to work?
+
+Detection:
+	Check /proc/scsi for "Vendor: HP", and NOT "Model: LOGICAL VOLUME". Take its scsi host number,
+		match against /proc/scsi/sg/devices (first column). The matched line number
+		is the one to use in /dev/sgN.
+	Run smartctl -i -d cciss,[0-127] /dev/cciss/cNd0
+		until "No such device or address" or "VALID ARGUMENTS ARE" is encountered in output.
+</pre> */
+inline std::string detect_drives_linux_hpsa(std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
+{
+	debug_out_info("app", DBG_FUNC_MSG << "Detecting drives behind HP RAID (hpsa/hpahcisr) controller(s)...\n");
+
+	std::vector< std::pair<int, std::string> > vendors_models;
+	std::string  error_msg = read_proc_scsi_scsi_file(vendors_models);
+	if (!error_msg.empty()) {
+		return error_msg;
+	}
+
+	std::vector< std::vector<int> > sg_entries;
+	error_msg = read_proc_scsi_sg_devices_file(sg_entries);
+	if (!error_msg.empty()) {
+		return error_msg;
+	}
+
+	hz::intrusive_ptr<CmdexSync> smartctl_ex = ex_factory->create_executor(ExecutorFactory::ExecutorSmartctl);
+
+	std::set<int> controller_hosts;
+
+	for (std::size_t i = 0; i < vendors_models.size(); ++i) {
+		if (!app_pcre_match("/Vendor: HP /i", vendors_models[i].second) || app_pcre_match("/LOGICAL VOLUME/i", vendors_models[i].second)) {
+			continue;  // not a supported controller, or a logical drive.
+		}
+		int host_num = vendors_models[i].first;
+		debug_out_dump("app", "Found HP controller in SCSI file, SCSI host " << host_num << ".\n");
+
+		// Skip additional adapters with the same host (not sure if this is needed, but it won't hurt).
+		if (controller_hosts.find(host_num) != controller_hosts.end()) {
+			debug_out_dump("app", "Skipping adapter with SCSI host " << host_num << ", host already found.\n");
+			continue;
+		}
+
+		controller_hosts.insert(host_num);
+
+		for (std::size_t sg_num = 0; sg_num < sg_entries.size(); ++sg_num) {
+			if (sg_entries[sg_num].size() < 3) {
+				// We need at least 3 columns in that file
+				continue;
+			}
+			if (sg_entries[sg_num][0] != host_num) {
+				// Different scsi host
+				continue;
+			}
+
+			std::string dev = std::string("/dev/sg") + hz::number_to_string(sg_num);
+
+			const int max_port = 127;
+			debug_out_dump("app", "Starting brute-force port scan on 0-" << max_port << " ports, device \"" << dev << "\".\n");
+
+			for (int port = 0; port <= max_port; ++port) {
+				StorageDeviceRefPtr drive(new StorageDevice(dev, std::string("cciss,") + hz::number_to_string(port)));
+
+				std::string local_error_msg = drive->fetch_basic_data_and_parse(smartctl_ex);
+				std::string output = drive->get_info_output();
+
+				if (app_pcre_match("/No such device or address/mi", output) || app_pcre_match("/VALID ARGUMENTS ARE/mi", output)) {
+					// We reached the controller port limit, or smartctl-supported port limit.
+					debug_out_dump("app", "Reached controller or smartctl port limit with port " << port << ", stopping port scan.\n");
+					break;
+				}
+				if (!local_error_msg.empty()) {
+					debug_out_info("app", "Smartctl returned with an error: " << local_error_msg << "\n");
+					debug_out_dump("app", "Skipping drive " << drive->get_device_with_type() << " due to smartctl error.\n");
+				} else {
+					drives.push_back(drive);
+					debug_out_info("app", "Added drive " << drive->get_device_with_type() << ".\n");
+				}
+			}
+
+			debug_out_dump("app", "Brute-force port scan finished.\n");
+		}
+	}
+
+	if (controller_hosts.empty()) {
+		debug_out_info("app", DBG_FUNC_MSG << "No hpsa/hpahcisr-specific entries found in Sg devices file.\n");
+	}
+
+	return error_msg;
+}
+
+
+
 }  // anon ns
 
 
@@ -808,6 +1071,8 @@ inline std::string detect_drives_linux_areca(std::vector<StorageDeviceRefPtr>& d
 
 std::string detect_drives_linux(std::vector<StorageDeviceRefPtr>& drives, ExecutorFactoryRefPtr ex_factory)
 {
+	clear_read_file_cache();
+
 	std::vector<std::string> error_msgs;
 	std::string error_msg;
 
@@ -837,6 +1102,16 @@ std::string detect_drives_linux(std::vector<StorageDeviceRefPtr>& drives, Execut
 		error_msgs.push_back(error_msg);
 	}
 
+	error_msg = detect_drives_linux_cciss(drives, ex_factory);
+	if (!error_msg.empty()) {
+		error_msgs.push_back(error_msg);
+	}
+
+	error_msg = detect_drives_linux_hpsa(drives, ex_factory);
+	if (!error_msg.empty()) {
+		error_msgs.push_back(error_msg);
+	}
+
 	return hz::string_join(error_msgs, "\n");
 }
 
